import psycopg2
import pandas as pd

from ovretl.db_utils.fetch_db_credentials import fetch_db_credentials


def select_from_table(entity: str, cluster="dev", schema="public") -> pd.DataFrame:
    """
    dbname is necessary to fetch Rates DB in dev & prod cluster
    :param entity:
    :param cluster: "dev" or "analytics"
    :param after:
    :param schema: "public" default, can be "station" for cluster "analytics"
    :return:
    """
    if cluster == "prod":
        raise ValueError("Cannot query from prod cluster")
    credentials = fetch_db_credentials(cluster=cluster)
    conn = psycopg2.connect(
        host=credentials["host"],
        user=credentials["username"],
        password=credentials["password"],
        dbname=credentials["dbname"],
        port=credentials["port"],
    )
    sql_query = pd.read_sql_query(
        """
    SELECT * FROM "{}"."{}"
    """.format(
            schema, entity
        ),
        conn,
    )

    df = pd.DataFrame(sql_query)
    return df



import pandas as pd


def merge_with_propositions_loads_then_with_shipments_loads(
    shipments_df: pd.DataFrame, propositions_loads_df: pd.DataFrame, shipments_loads_df: pd.DataFrame,
) -> pd.DataFrame:
    shipments_df = pd.merge(
        shipments_df,
        propositions_loads_df[
            [
                "total_number",
                "total_volume_in_cbm",
                "total_weight_in_kg",
                "taxable_weight_in_kg",
                "weight_measurable",
                "hazardous",
                "lithium",
                "non_stackable",
                "proposition_id",
            ]
        ],
        how="left",
        on="proposition_id",
    )

    shipments_df_with_total_weight = shipments_df[~shipments_df["total_weight_in_kg"].isnull()]
    shipments_df_without_total_weight = shipments_df[shipments_df["total_weight_in_kg"].isnull()].drop(
        [
            "total_number",
            "total_weight_in_kg",
            "total_volume_in_cbm",
            "taxable_weight_in_kg",
            "weight_measurable",
            "hazardous",
            "lithium",
            "non_stackable",
        ],
        axis=1,
    )

    shipments_df_without_total_weight = pd.merge(
        shipments_df_without_total_weight,
        shipments_loads_df[
            [
                "total_number",
                "total_volume_in_cbm",
                "total_weight_in_kg",
                "taxable_weight_in_kg",
                "weight_measurable",
                "hazardous",
                "magnetic",
                "lithium",
                "refrigerated",
                "non_stackable",
                "shipment_id",
            ]
        ],
        how="left",
        on="shipment_id",
    )

    return (
        pd.concat([shipments_df_with_total_weight, shipments_df_without_total_weight], sort=False,)
        .reset_index(drop=True)
        .rename(
            columns={
                "hazardous": "loads_hazardous",
                "magnetic": "loads_magnetic",
                "lithium": "loads_lithium",
                "refrigerated": "loads_refrigerated",
            }
        )
    )


def split_propositions_shipments_loads(loads_df: pd.DataFrame):
    propositions_loads_df = loads_df[~loads_df["proposition_id"].isnull()]
    shipments_loads_df = loads_df[~loads_df["shipment_id"].isnull()]
    return propositions_loads_df, shipments_loads_df


def merge_shipments_with_loads(shipments_df: pd.DataFrame, loads_df: pd.DataFrame, drop_duplicate_key="shipment_id"):
    (propositions_loads_df, shipments_loads_df,) = split_propositions_shipments_loads(loads_df)
    shipments_df = merge_with_propositions_loads_then_with_shipments_loads(
        shipments_df=shipments_df, propositions_loads_df=propositions_loads_df, shipments_loads_df=shipments_loads_df,
    )
    shipments_df = shipments_df.sort_values(by=["total_weight_in_kg", "total_volume_in_cbm"], ascending=False)
    shipments_df = shipments_df[
        (~shipments_df.duplicated(subset=[drop_duplicate_key])) | (shipments_df[drop_duplicate_key].isnull())
    ]  # drop duplicates without dropping nan
    return shipments_df



import psycopg2
import pandas as pd
from ovretl.db_utils.fetch_db_credentials import fetch_db_credentials


def execute_complex_insert(df: pd.DataFrame, stage="dev"):
    values_to_insert = [tuple(x) for x in df.values]
    columns = ",".join(df.columns)
    credentials = fetch_db_credentials(stage)
    conn = psycopg2.connect(
        host=credentials["host"],
        user=credentials["username"],
        password=credentials["password"],
        dbname="Rates",
        port=6543,
    )
    cur = conn.cursor()
    from psycopg2.extras import execute_values

    execute_values(
        cur,
        """INSERT INTO "RateMeta" (
                  {}
                  ) VALUES %s ON CONFLICT ON CONSTRAINT {}
                  DO UPDATE SET
                  "updated_at" = NOW(),
                  "margin_type" = EXCLUDED."margin_type",
                  "margin_value" = EXCLUDED."margin_value"
                   """.format(
            columns, "unicity_on_kronos_name"
        ),
        values_to_insert,
    )
    cur.close()
    conn.commit()
    conn.close()
    print("Inserted {} rows.".format(cur.rowcount))



import pandas as pd


def join_employees_associations_to_employees_name(
    employees_df: pd.DataFrame, employee_shipment_associations_df: pd.DataFrame
):
    return pd.merge(
        employees_df,
        employee_shipment_associations_df[["employee_id", "shipment_id", "role"]],
        left_on="id",
        right_on="employee_id",
    )



import pandas as pd


def merge_with_propositions_loads(
    propositions_df: pd.DataFrame, propositions_loads_df: pd.DataFrame) -> pd.DataFrame:
    propositions_df = pd.merge(
        propositions_df,
        propositions_loads_df[
            [
                "total_number",
                "total_volume_in_cbm",
                "total_weight_in_kg",
                "taxable_weight_in_kg",
                "weight_measurable",
                "hazardous",
                "lithium",
                "non_stackable",
                "proposition_id",
            ]
        ],
        how="left",
        on="proposition_id",
    ).reset_index(drop=True).rename(columns={"hazardous": "loads_hazardous", "magnetic": "loads_magnetic",
                                             "lithium": "loads_lithium", "refrigerated": "loads_refrigerated"})
    return propositions_df


def merge_propositions_with_loads(
        propositions_df: pd.DataFrame, propositions_loads_df: pd.DataFrame, drop_duplicate_key="proposition_id"
):
    propositions_df = merge_with_propositions_loads(
        propositions_df=propositions_df, propositions_loads_df=propositions_loads_df
    )
    propositions_df = propositions_df.sort_values(by=["total_weight_in_kg", "total_volume_in_cbm"], ascending=False)
    propositions_df = propositions_df[
        (~propositions_df.duplicated(subset=[drop_duplicate_key])) | (propositions_df[drop_duplicate_key].isnull())
    ]  # drop duplicates without dropping nan
    return propositions_df



from typing import Optional
import datetime
import numpy as np
import pandas as pd
import pytz

START_TIME_PARIS = datetime.time(9, 0)
END_TIME_PARIS = datetime.time(18, 0)
TOTAL_WORKED_HOURS = END_TIME_PARIS.hour - START_TIME_PARIS.hour


def calculate_time_difference_within_day_in_hours(date_1: datetime.time, date_2: datetime.time):
    date_1 = datetime.datetime.combine(datetime.date.today(), date_1)  # type: ignore
    date_2 = datetime.datetime.combine(datetime.date.today(), date_2)  # type: ignore
    datetime_difference = date_2 - date_1  # type: ignore
    return max(0.0, round(datetime_difference.total_seconds() / 3600, 2))


def is_weekday(date: datetime.datetime) -> bool:
    if date.isoweekday() in [6, 7]:
        return False
    return True


def calculate_business_hours_delta(date_1: Optional[datetime.datetime], date_2: Optional[datetime.datetime]):
    """
    Calculate time difference in worked hours, based on Mo-Fr, working from 9:00 to 18:00 in the Paris office
    Careful: in this function we only consider the Paris office case (to change when we'll deal with other timezones)
    :param date_1:
    :param date_2:
    :return:
    """
    if date_1 is None or date_2 is None or pd.isna(date_1) or pd.isna(date_2):
        return None
    # we need to convert the UTC times we pull from the DB into Paris time
    date_1 = pytz.timezone("UTC").localize(date_1, is_dst=None).astimezone('Europe/Paris').replace(tzinfo=None)
    date_2 = pytz.timezone("UTC").localize(date_2, is_dst=None).astimezone('Europe/Paris').replace(tzinfo=None)
    business_days_diff = np.busday_count(date_1.strftime("%Y-%m-%d"), date_2.strftime("%Y-%m-%d"))
    if business_days_diff == 0 and date_1.date() == date_2.date():
        return calculate_time_difference_within_day_in_hours(date_1.time(), date_2.time())
    if business_days_diff < 0:
        return np.nan
    if not np.is_busday(date_1.strftime("%Y-%m-%d")):
        business_days_diff += 1
    first_day_remaining_h = (
        calculate_time_difference_within_day_in_hours(date_1.time(), END_TIME_PARIS) if is_weekday(date_1) else 0
    )
    second_day_remaining_h = (
        calculate_time_difference_within_day_in_hours(START_TIME_PARIS, date_2.time()) if is_weekday(date_2) else 0
    )
    return max(0, (business_days_diff - 1)) * TOTAL_WORKED_HOURS + first_day_remaining_h + second_day_remaining_h



from ovretl.prices_utils.sum_multiple_billings_prices import sum_multiple_billings_prices
import pandas as pd


def add_initial_category_prices_to_shipments(
    shipments_with_prices: pd.DataFrame, prices_propositions_by_category_df: pd.DataFrame,
) -> pd.DataFrame:
    initial_prices = (
        prices_propositions_by_category_df.drop(["count"], axis=1, errors="ignore")
        .dropna(subset=["purchase_price_in_eur", "price_in_eur", "margin_price_in_eur", "proposition_id",], how="all",)
        .rename(
            columns={
                "purchase_price_in_eur": "initial_purchase_price_in_eur",
                "price_in_eur": "initial_price_in_eur",
                "margin_price_in_eur": "initial_margin_price_in_eur",
                "proposition_id": "initial_proposition_id",
            }
        )
    )

    shipments_with_prices = pd.merge(
        shipments_with_prices,
        initial_prices[
            [
                "initial_proposition_id",
                "category",
                "initial_margin_price_in_eur",
                "initial_price_in_eur",
                "initial_purchase_price_in_eur",
            ]
        ],
        on=("initial_proposition_id", "category"),
        how="left",
    )
    return shipments_with_prices


def add_category_prices_to_shipments(
    shipments_with_prices_ids: pd.Series,
    prices_final_check_by_category_df: pd.DataFrame,
    prices_billings_by_category_df: pd.DataFrame,
    prices_propositions_by_category_df: pd.DataFrame,
) -> pd.DataFrame:
    shipments_with_prices = pd.merge(
        shipments_with_prices_ids,
        prices_final_check_by_category_df.dropna(subset=["price_in_eur", "final_check_id"], how="any"),
        on="final_check_id",
        how="left",
    )

    shipments_with_prices_full = shipments_with_prices[~shipments_with_prices["price_in_eur"].isnull()]
    shipments_with_prices_uncomplete = shipments_with_prices[shipments_with_prices["price_in_eur"].isnull()].drop(
        [
            "category",
            "price_in_eur",
            "vat_price_in_eur",
            "margin_price_in_eur",
            "purchase_price_in_eur",
            "prices_origin",
        ],
        axis=1,
    )

    shipments_with_prices_uncomplete = pd.merge(
        shipments_with_prices_uncomplete,
        prices_billings_by_category_df.dropna(subset=["price_in_eur", "billing_id"], how="any"),
        on="billing_id",
        how="left",
    )

    shipments_with_prices = pd.concat([shipments_with_prices_full, shipments_with_prices_uncomplete], sort=False)
    shipments_with_prices_full = shipments_with_prices[~shipments_with_prices["price_in_eur"].isnull()]
    shipments_with_prices_uncomplete = shipments_with_prices[shipments_with_prices["price_in_eur"].isnull()].drop(
        [
            "category",
            "price_in_eur",
            "vat_price_in_eur",
            "margin_price_in_eur",
            "purchase_price_in_eur",
            "prices_origin",
        ],
        axis=1,
    )

    shipments_with_prices_uncomplete = pd.merge(
        shipments_with_prices_uncomplete,
        prices_propositions_by_category_df.dropna(subset=["price_in_eur", "proposition_id"], how="any"),
        on="proposition_id",
        how="left",
    )

    shipments_with_prices = pd.concat([shipments_with_prices_full, shipments_with_prices_uncomplete], sort=False)

    shipments_with_prices = add_initial_category_prices_to_shipments(
        shipments_with_prices=shipments_with_prices,
        prices_propositions_by_category_df=prices_propositions_by_category_df,
    )

    return sum_multiple_billings_prices(shipments_with_prices=shipments_with_prices)



import os

import pandas as pd
from ovretl.transit_times.preprocess_transit_times import preprocess_transit_times
from pandas.util.testing import assert_frame_equal

cwd = os.path.dirname(__file__)


def test_preprocess_transit_times():
    raw_df = pd.read_csv(os.path.join(cwd, "transit_times_data.csv"))
    result = preprocess_transit_times(raw_df, 20)
    result_should_be = pd.read_csv(os.path.join(cwd, "preprocessed_transit_times_data.csv"))
    assert_frame_equal(result.reset_index(drop=True), result_should_be.reset_index(drop=True))



import pandas as pd

from ovretl.performances_utils.calculate_business_days_delta import calculate_business_days_delta
from ovretl.performances_utils.calculate_business_hours_delta import calculate_business_hours_delta

QUOTATION_ASKED = "quotation_asked"
QUOTATION_CONSIDERED = "request_considered"
QUOTATION_PURCHASE_READY = "quotation_purchase_ready"
QUOTATION_SENT = "propositions_received"
QUOTATION_ACCEPTED = "quotation_accepted"
SHIPMENT_BOOKED = "shipment_booked"
SHIPMENT_ARRIVED = "shipment_arrived"
SHIPMENT_FINISHED = "shipment_finished"
BILLING_AVAILABLE = "billing_available"
BILLING_PAID = "billing_paid"


def drop_not_arrived_shipment_groups(shipment_events_df: pd.DataFrame):
    shipment_events_df = shipment_events_df.sort_values(by=["date"], ascending=False).reset_index(drop=True)
    if shipment_events_df.loc[0, "date_type"] == "estimated":
        return False
    return True


def extract_arrived_date_from_events(events_shipment_df: pd.DataFrame):
    events_shipment_df = events_shipment_df.dropna(subset=["shipment_id", "date"], how="any").drop(
        ["created_at"], axis=1
    )
    events_shipment_df = events_shipment_df.groupby("shipment_id").filter(drop_not_arrived_shipment_groups)
    events_shipment_df = events_shipment_df.sort_values(by=["shipment_id", "date"], ascending=False)
    events_shipment_df = events_shipment_df.drop_duplicates(subset=["shipment_id"])
    events_shipment_df.loc[:, "header"] = "shipment_arrived"
    events_shipment_df = events_shipment_df.rename(columns={"date": "created_at"})
    return events_shipment_df[["created_at", "header", "shipment_id"]]


def extract_finished_date_from_shipments(shipments_df: pd.DataFrame):
    shipments_df = shipments_df.dropna(subset=["finished_date"]).drop(["created_at"], axis=1)
    shipments_df.loc[:, "header"] = "shipment_finished"
    shipments_df = shipments_df.rename(columns={"finished_date": "created_at", "id": "shipment_id"})
    return shipments_df[["created_at", "header", "shipment_id"]]


def concatenate_activities(
    events_shipment_df: pd.DataFrame, activities_clean_df: pd.DataFrame, shipments_df: pd.DataFrame,
):
    activities_clean_df = pd.concat([activities_clean_df, events_shipment_df, shipments_df], sort=False)
    activities_clean_df.loc[:, "created_at"] = pd.to_datetime(activities_clean_df["created_at"]).apply(
        lambda t: t.replace(tzinfo=None)
    )
    activities_clean_df = activities_clean_df.sort_values(by=["shipment_id", "header", "created_at"], ascending=True)
    activities_clean_df = activities_clean_df.drop_duplicates(subset=["shipment_id", "header"])

    activities_clean_df = activities_clean_df[
        activities_clean_df["header"].isin(
            [
                QUOTATION_ASKED,
                QUOTATION_CONSIDERED,
                QUOTATION_PURCHASE_READY,
                QUOTATION_SENT,
                QUOTATION_ACCEPTED,
                SHIPMENT_BOOKED,
                SHIPMENT_ARRIVED,
                SHIPMENT_FINISHED,
                BILLING_AVAILABLE,
                BILLING_PAID,
            ]
        )
    ]
    activities_clean_df = pd.pivot_table(
        activities_clean_df, values="created_at", index=["shipment_id"], columns=["header"], aggfunc="first",
    )
    return activities_clean_df.reset_index()


def calculate_steps_time(activities_df: pd.DataFrame):
    activities_df.loc[:, "quotation_asked_considered_hours"] = activities_df.apply(
        lambda s: calculate_business_hours_delta(s[QUOTATION_ASKED], s[QUOTATION_CONSIDERED]), axis=1,
    )
    activities_df.loc[:, "quotation_considered_purchase_ready_hours"] = activities_df.apply(
        lambda s: calculate_business_hours_delta(s[QUOTATION_CONSIDERED], s[QUOTATION_PURCHASE_READY]), axis=1,
    )
    activities_df.loc[:, "quotation_purchase_ready_sent_hours"] = activities_df.apply(
        lambda s: calculate_business_hours_delta(s[QUOTATION_PURCHASE_READY], s[QUOTATION_SENT]), axis=1,
    )
    activities_df.loc[:, "quotation_asked_sent_hours"] = activities_df.apply(
        lambda s: calculate_business_hours_delta(s[QUOTATION_ASKED], s[QUOTATION_SENT]), axis=1,
    )
    activities_df.loc[:, "quotation_accepted_shipment_booked_days"] = activities_df.apply(
        lambda s: calculate_business_days_delta(s[QUOTATION_ACCEPTED], s[SHIPMENT_BOOKED]), axis=1,
    )
    activities_df.loc[:, "shipment_arrived_shipment_finished_days"] = activities_df.apply(
        lambda s: calculate_business_days_delta(s[SHIPMENT_ARRIVED], s[SHIPMENT_FINISHED]), axis=1,
    )
    activities_df.loc[:, "shipment_arrived_billing_available_days"] = activities_df.apply(
        lambda s: calculate_business_days_delta(s[SHIPMENT_ARRIVED], s[BILLING_AVAILABLE]), axis=1,
    )
    activities_df.loc[:, "billing_available_paid_days"] = activities_df.apply(
        lambda s: calculate_business_days_delta(s[BILLING_AVAILABLE], s[BILLING_PAID]), axis=1,
    )
    return activities_df


def calculate_shipments_treatment_steps_time(
    events_shipment_df: pd.DataFrame, activities_clean_df: pd.DataFrame, shipments_df: pd.DataFrame,
):
    events_shipment_df = extract_arrived_date_from_events(events_shipment_df=events_shipment_df)
    activities_clean_df = activities_clean_df.dropna(subset=["shipment_id"])[
        ["created_at", "header", "shipment_id", "employee_id"]
    ]
    shipments_df = extract_finished_date_from_shipments(shipments_df=shipments_df)

    activities_clean_df = concatenate_activities(
        events_shipment_df=events_shipment_df, activities_clean_df=activities_clean_df, shipments_df=shipments_df,
    )
    return calculate_steps_time(activities_df=activities_clean_df).drop(["employee_id"], axis=1, errors="ignore")



from ovretl.performances_utils.calculate_business_hours_delta import calculate_business_hours_delta
import datetime
import pandas as pd


def test_calculate_business_hours_delta():
    date_1 = pd.to_datetime(datetime.datetime(2020, 6, 26, 15, 45))
    date_2 = pd.to_datetime(datetime.datetime(2020, 6, 29, 9, 30))
    assert calculate_business_hours_delta(date_1, date_2) == 2.75

    date_3 = pd.to_datetime(datetime.datetime(2020, 6, 25, 15, 30))
    date_4 = pd.to_datetime(datetime.datetime(2020, 6, 26, 9, 30))
    assert calculate_business_hours_delta(date_3, date_4) == 3.0

    date_5 = pd.to_datetime(datetime.datetime(2020, 6, 25, 9, 30))
    date_6 = pd.to_datetime(datetime.datetime(2020, 6, 26, 15, 30))
    assert calculate_business_hours_delta(date_5, date_6) == 15.0

    date_7 = pd.to_datetime(datetime.datetime(2020, 6, 25, 9, 45))
    date_8 = pd.to_datetime(datetime.datetime(2020, 6, 25, 15, 30))
    assert calculate_business_hours_delta(date_7, date_8) == 5.75

    date_9 = pd.to_datetime(datetime.datetime(2020, 6, 25, 15, 30))
    date_10 = pd.to_datetime(datetime.datetime(2020, 6, 27, 19, 30))
    assert calculate_business_hours_delta(date_9, date_10) == 9.5

    date_11 = pd.to_datetime(datetime.datetime(2020, 6, 26, 15, 30))
    date_12 = pd.to_datetime(datetime.datetime(2020, 6, 30, 9, 30))
    assert calculate_business_hours_delta(date_11, date_12) == 12.0

    date_13 = pd.to_datetime(datetime.datetime(2020, 6, 28, 20, 30))
    date_14 = pd.to_datetime(datetime.datetime(2020, 6, 29, 10, 30))
    assert calculate_business_hours_delta(date_13, date_14) == 3.5

    date_15 = pd.to_datetime(datetime.datetime(2020, 8, 22, 17, 30))
    date_16 = pd.to_datetime(datetime.datetime(2020, 8, 25, 8, 30))
    assert calculate_business_hours_delta(date_15, date_16) == 10.5



import pandas as pd


def match_kronos_propositions_and_shipments(
    kronos_propositions_df: pd.DataFrame, shipments_all_propositions_df: pd.DataFrame
):
    kronos_propositions_df = kronos_propositions_df.rename(columns={"id": "kronos_proposition_id"})
    kronos_propositions_df = kronos_propositions_df.dropna(subset=["shipment_id"])

    # Merge on kronos_proposition_id to get kronos_action (new system: table propositions has kronos_proposition_id)
    kronos_propositions_df = pd.merge(
        kronos_propositions_df,
        shipments_all_propositions_df[
            ["kronos_proposition_id", "proposition_id", "kronos_action", "foresea_name"]
        ].dropna(subset=["kronos_proposition_id"]),
        how="left",
        on="kronos_proposition_id",
    )

    kronos_propositions_df_with_foresea_name = (
        kronos_propositions_df[~kronos_propositions_df["foresea_name"].isnull()]
        .drop(["proposition_id_x"], axis=1)
        .rename(columns={"proposition_id_y": "proposition_id"})
    )
    kronos_propositions_df_without_foresea_name = (
        kronos_propositions_df[kronos_propositions_df["foresea_name"].isnull()]
        .drop(["foresea_name", "proposition_id_y", "kronos_action"], axis=1)
        .rename(columns={"proposition_id_x": "proposition_id"})
    )

    # Merge on proposition_id for remaining kronos_propositions (old system: table kronos_proposition_id has proposition_id)
    kronos_propositions_df_without_foresea_name = pd.merge(
        kronos_propositions_df_without_foresea_name,
        shipments_all_propositions_df[["proposition_id", "kronos_action"]].dropna(subset=["proposition_id"]),
        how="left",
        on="proposition_id",
    )

    # Concatenate
    kronos_propositions_df = pd.concat(
        [kronos_propositions_df_with_foresea_name, kronos_propositions_df_without_foresea_name], sort=False
    ).reset_index(drop=True)

    # Sort with shipment_id, then proposition_id to put most relevant Kronos propositions first
    kronos_propositions_df = kronos_propositions_df.sort_values(
        by=["shipment_id", "proposition_id", "created_at"], ascending=False,
    )
    kronos_propositions_df = kronos_propositions_df.drop_duplicates(subset=["shipment_id"])
    return kronos_propositions_df.drop(["foresea_name"], axis=1)
